"""URL fragment normalization."""

from __future__ import annotations

from .tools import quote, unquote


def normalize_fragment(fragment: str) -> str:
    """Normalize fragment part of the url.

    Params:
        fragment : string : url fragment, e.g., 'fragment'

    Returns:
        string : normalized fragment data.

    Notes:
        According to RFC 3986, the following characters are allowed in a fragment:
        fragment    = *( pchar / "/" / "?" )
        pchar       = unreserved / pct-encoded / sub-delims / ":" / "@"
        unreserved  = ALPHA / DIGIT / "-" / "." / "_" / "~"
        sub-delims  = "!" / "$" / "&" / "'" / "(" / ")" / "*" / "+" / "," / ";" / "="
        We specifically allow "~" and "=" as safe characters during normalization.
        Other sub-delimiters could potentially be added to the `safe` list if needed.

    """
    return quote(unquote(fragment), safe="~=")



#!/usr/bin/env python
"""Command line interface for url-normalize."""

import argparse
import sys
from importlib.metadata import version

from .url_normalize import url_normalize


def main() -> None:
    """Parse arguments and run url_normalize."""
    parser = argparse.ArgumentParser(description="Normalize a URL.")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {version('url-normalize')}",
    )
    parser.add_argument("url", help="The URL to normalize.")
    parser.add_argument(
        "-c",
        "--charset",
        default="utf-8",
        help="The charset of the URL. Default: utf-8",
    )
    parser.add_argument(
        "-s",
        "--default-scheme",
        default="https",
        help="The default scheme to use if missing. Default: https",
    )
    parser.add_argument(
        "-f",
        "--filter-params",
        action="store_true",
        help="Filter common tracking parameters.",
    )
    parser.add_argument(
        "-d",
        "--default-domain",
        type=str,
        help="Default domain to use for absolute paths (starting with '/').",
    )
    parser.add_argument(
        "-p",
        "--param-allowlist",
        type=str,
        help="Comma-separated list of query parameters to allow (e.g., 'q,id').",
    )

    args = parser.parse_args()

    allowlist = args.param_allowlist.split(",") if args.param_allowlist else None

    try:
        normalized_url = url_normalize(
            args.url,
            charset=args.charset,
            default_scheme=args.default_scheme,
            default_domain=args.default_domain,
            filter_params=args.filter_params,
            param_allowlist=allowlist,
        )
    except Exception as e:  # noqa: BLE001
        print(f"Error normalizing URL: {e}", file=sys.stderr)  # noqa: T201
        sys.exit(1)
    else:
        print(normalized_url)  # noqa: T201


if __name__ == "__main__":
    main()  # pragma: no cover



"""URL normalize main module.

Copyright (c) 2020 Nikolay Panov
This module is part of url-normalize package and is released under the MIT License:
https://opensource.org/licenses/MIT

"""

from __future__ import annotations

from .generic_url_cleanup import generic_url_cleanup
from .normalize_fragment import normalize_fragment
from .normalize_host import DEFAULT_CHARSET, normalize_host
from .normalize_path import normalize_path
from .normalize_port import normalize_port
from .normalize_query import normalize_query
from .normalize_scheme import DEFAULT_SCHEME, normalize_scheme
from .normalize_userinfo import normalize_userinfo
from .provide_url_domain import provide_url_domain
from .provide_url_scheme import provide_url_scheme
from .tools import deconstruct_url, reconstruct_url


def url_normalize(  # noqa: PLR0913
    url: str | None,
    *,  # Force keyword-only arguments
    charset: str = DEFAULT_CHARSET,
    default_scheme: str = DEFAULT_SCHEME,
    default_domain: str | None = None,
    filter_params: bool = False,
    param_allowlist: dict | list | None = None,
) -> str | None:
    """URI normalization routine.

    Sometimes you get an URL by a user that just isn't a real
    URL because it contains unsafe characters like ' ' and so on.
    This function can fix some of the problems in a similar way
    browsers handle data entered by the user:

    >>> url_normalize('http://de.wikipedia.org/wiki/Elf (Begriffsklärung)')
    'http://de.wikipedia.org/wiki/Elf%20%28Begriffskl%C3%A4rung%29'

    Params:
        url : str | None : URL to normalize
        charset : str : optional
            The target charset for the URL if the url was given as unicode string
        default_scheme : str : default scheme to use if none present
        default_domain : str | None : optional
            Default domain to use for absolute paths (starting with '/')
        filter_params : bool : optional
            Whether to filter non-allowlisted parameters (False by default)
        param_allowlist : dict | list | None : optional
            Override for the parameter allowlist

    Returns:
        str | None : a normalized url

    """
    if not url:
        return url
    url = provide_url_domain(url, default_domain)
    url = provide_url_scheme(url, default_scheme)
    url = generic_url_cleanup(url)
    url_elements = deconstruct_url(url)
    url_elements = url_elements._replace(
        scheme=normalize_scheme(url_elements.scheme),
        userinfo=normalize_userinfo(url_elements.userinfo),
        host=normalize_host(url_elements.host, charset),
        query=normalize_query(
            url_elements.query,
            host=url_elements.host,
            filter_params=filter_params,
            param_allowlist=param_allowlist,
        ),
        fragment=normalize_fragment(url_elements.fragment),
    )
    url_elements = url_elements._replace(
        port=normalize_port(url_elements.port, url_elements.scheme),
        path=normalize_path(url_elements.path, url_elements.scheme),
    )
    return reconstruct_url(url_elements)



# Copyright (c) 2023. All rights reserved.
"""URL query parameter allowlist module."""

from __future__ import annotations

DEFAULT_ALLOWLIST = {
    "google.com": ["q", "ie"],
    "baidu.com": ["wd", "ie"],
    "bing.com": ["q"],
    "youtube.com": ["v", "search_query"],
}


def get_allowed_params(
    host: str | None = None,
    allowlist: dict | list | None = None,
) -> set[str]:
    """Get allowed parameters for a given domain.

    Params:
        host: Domain name to check (e.g. 'google.com')
        allowlist: Optional override for default allowlist
            If provided as a list, it will be used as is.
            If provided as a dictionary, it should map domain names to
            lists of allowed parameters.
            If None, the default allowlist will be used.

    Returns:
        Set of allowed parameter names for the domain

    """
    if isinstance(allowlist, list):
        return set(allowlist)

    if not host:
        return set()

    # Normalize host by removing www and port
    domain = host.lower()
    if domain.startswith("www."):
        domain = domain[4:]
    domain = domain.split(":")[0]

    # Use default allowlist if none provided
    if allowlist is None:
        allowlist = DEFAULT_ALLOWLIST

    # Return allowed parameters for the domain, or an empty set if not found
    return set(allowlist.get(domain, []))



"""URL path normalization."""

from __future__ import annotations

from .tools import quote, unquote


def normalize_path(path: str, scheme: str) -> str:
    """Normalize path part of the url.

    Remove mention of default path number

    Params:
        path : string : url path, e.g., '/section/page.html'
        scheme : string : url scheme, e.g., 'http'

    Returns:
        string : normalized path data.

    """
    # Only perform percent-encoding where it is essential.
    # Always use uppercase A-through-F characters when percent-encoding.
    # All portions of the URI must be utf-8 encoded NFC from Unicode strings
    path = quote(unquote(path), "~:/#[]@!$&'()*+,;=")
    # Prevent dot-segments appearing in non-relative URI paths.
    if scheme in {"", "http", "https", "ftp", "file"}:
        output: list[str] = []
        for part in path.split("/"):
            if part == "":
                if not output:
                    output.append(part)
            elif part == ".":
                pass
            elif part == "..":
                if len(output) > 1:
                    output.pop()
            else:
                output.append(part)
        # The part variable is used in the final check
        last_part = part
        if last_part in {"", ".", ".."}:
            output.append("")
        path = "/".join(output)
    # For schemes that define an empty path to be equivalent to a path of "/",
    # use "/".
    if not path and scheme in {"http", "https", "ftp", "file"}:
        path = "/"
    return path



"""URL generic cleanup operations."""

from __future__ import annotations


def generic_url_cleanup(url: str) -> str:
    """Cleanup the URL from unnecessary data and convert to final form.

    Converts shebang urls to final form, removed unnecessary data from the url.

    Params:
        url : string : the URL

    Returns:
        string : update url

    """
    url = url.replace("#!", "?_escaped_fragment_=")
    return url.rstrip("&? ")



"""URL host normalization."""

from __future__ import annotations

import idna

from .tools import force_unicode

DEFAULT_CHARSET = "utf-8"


def normalize_host(host: str, charset: str = DEFAULT_CHARSET) -> str:
    """Normalize host part of the url.

    Lowercase and strip of final dot.
    Also, handle IDN domains using IDNA2008 with UTS46 transitional processing.

    Params:
        host : string : url host, e.g., 'site.com'
        charset : string : encoding charset

    Returns:
        string : normalized host data.

    """
    host = force_unicode(host, charset)
    host = host.lower()
    host = host.strip(".")

    # Split domain into parts to handle each label separately
    parts = host.split(".")
    try:
        # Process each label separately to handle mixed unicode/ascii domains
        parts = [
            idna.encode(p, uts46=True, transitional=True).decode(charset)
            for p in parts
            if p
        ]
        return ".".join(parts)
    except idna.IDNAError:
        # Fallback to direct encoding if IDNA2008 processing fails
        return host.encode("idna").decode(charset)
